<?php

namespace Drupal\abtestui\Controller;

use Drupal\abtestui\Service\TestStorage;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class ABTestUIDefaultController.
 *
 * @package Drupal\abtestui\Controller
 */
class ABTestUIDefaultController extends ControllerBase {

  /**
   * THe custom abtestui test storage.
   *
   * @var \Drupal\abtestui\Service\TestStorage
   */
  private $testStorage;

  private $renderer;

  private $dateFormatter;

  private $analyticsConfig;

  private $abtestuiPath;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('abtestui.test_storage'),
      $container->get('renderer'),
      $container->get('date.formatter'),
      $container->get('config.factory')
    );
  }

  public function __construct(
    TestStorage $testStorage,
    RendererInterface $renderer,
    DateFormatterInterface $dateFormatter,
    ConfigFactoryInterface $configFactory
  ) {
    $this->testStorage = $testStorage;
    $this->renderer = $renderer;
    $this->dateFormatter = $dateFormatter;
    $this->analyticsConfig = $configFactory->get('google_analytics.settings');
    $this->abtestuiPath = drupal_get_path('module', 'abtestui');
  }

  /**
   * Return the list page.
   *
   * @return array
   *   Render array.
   */
  public function testList() {
    $tests = $this->testStorage->loadMultiple();
    $alphabets = range('A', 'Z');

    $output = [
      '#type' => 'container',
      '#attributes' => [
        'class' => [
          'test-content',
        ],
      ],
    ];
    $rows = [];

    foreach ($tests as $test) {
      $test_name = [
        'test_name_col' => [
          '#type' => 'container',
          '#attributes' => [
            'class' => [
              'test-name-col',
            ],
          ],
          'test_name' => [
            '#title' => $test['name'],
            '#type' => 'link',
            '#url' => Url::fromRoute('abtestui.test_edit_form', ['ab_test_id' => $test['tid']]),
          ],
        ],
      ];
      $base_name = [
        'base_url' => [
          '#type' => 'container',
          '#attributes' => [
            'class' => [
              'base-url',
              'ab-variations',
            ],
          ],
          'list_counter' => [
            '#type' => 'container',
            '#attributes' => [
              'class' => [
                'list-counter',
              ],
            ],
            'markup' => [
              '#markup' => $alphabets[0],
            ],
          ],
          'variation_name' => [
            '#type' => 'container',
            '#attributes' => [
              'class' => [
                'variation-name',
              ],
            ],
            'markup' => [
              '#markup' => t('Base URL') . '<a href="/' . $this->abtestuiPath . '/help/base-url.html" class="help-modal"><img width="20" height="20" src="/' . $this->abtestuiPath . '/img/help.svg"></a>',
            ],
          ],
        ],
      ];

      $base_url = [
        'variation_url' => [
          '#type' => 'container',
          '#attributes' => [
            'class' => [
              'variation-url',
              'base-url',
            ],
          ],
          'link' => [
            '#title' => $test['base_url'],
            '#type' => 'link',
            '#url' => Url::fromUri($test['base_url']),
          ],
        ],
      ];

      $i = 1;
      $inner_table_rows = [];
      $inner_table_rows[] = [
        $this->renderer->render($test_name),
        $this->renderer->render($base_name),
        $this->renderer->render($base_url),
      ];
      /** @var array $variation */
      foreach ($test['variations'] as $variation) {
        $variation_name = [];
        $prefield = [];
        if ($i == 1) {
          if ($test['analytics_url']) {
            $prefield['analitycs_url'] =[
              '#type' => 'container',
              '#attributes' => [
                'class' => [
                  'test-ga-link',
                ],
              ],
              'link' => [
                '#title'  => t('Analytics Results'),
                '#type'   => 'link',
                '#url'    => Url::fromUri($test['analytics_url']),
              ],
            ];
          }
        }

        $variation_name['variation_' . $i] = [
          '#type'          => 'container',
          '#attributes'    => [
            'class' => [
              'ab-variations',
            ],
          ],
          'list_counter'   => [
            '#type'       => 'container',
            '#attributes' => [
              'class' => [
                'list-counter',
              ],
            ],
            'markup'      => [
              '#markup' => "$alphabets[$i]",
            ],
          ],
          'variation_name' => [
            '#type'       => 'container',
            '#attributes' => [
              'class' => [
                'variation-name',
              ],
            ],
            'markup'      => [
              '#markup' => t('@variation_name', [
                '@variation_name' => $variation['name'],
              ]),
            ],
          ],
        ];

        $variation_url = [
          'variation_url'  => [
            '#type'       => 'container',
            '#attributes' => [
              'class' => [
                'variation-url',
              ],
            ],
            'link'        => [
              '#title' => $variation['url'],
              '#type'  => 'link',
              '#url'   => Url::fromUri($variation['url']),
            ],
          ],
        ];

        $i++;
        $inner_table_rows[] = [
          $this->renderer->render($prefield),
          $this->renderer->render($variation_name),
          $this->renderer->render($variation_url),
        ];
      }

      $inner_table = [
        '#type' => 'table',
        '#header' => NULL,
        '#rows' => $inner_table_rows,
        '#attributes' => [
          'class' => [
            'ab-test-inner-table',
          ],
        ],
      ];

      $rows[] = [
        $this->renderer->render($inner_table),
      ];
    }

    if (empty($this->analyticsConfig->get('account'))) {
      $ga_account_id = '<a href="' . Url::fromRoute('google_analytics.admin_settings_form')->toString() . '">' . $this->t('Set up') . '</a>';
    }
    else {
      $ga_account_id = '#' . $this->analyticsConfig->get('account');
    }

    $output['info_bar'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => [
          'info-bar',
        ],
      ],
      'title' => [
        '#markup' => '<h2>' . t('A/B Testing Tool') . '</h2>',
      ],
      'ab_test_tool_help' => [
        '#type' => 'container',
        '#attributes' => [
          'class' => [
            'ab-test-tool-help',
          ],
        ],
        'link' => [
          '#title' => t('How does this tool work?'),
          '#type' => 'link',
          '#url' => Url::fromUri('base:/' . $this->abtestuiPath . '/help/ab-test-tool-help.html'),
          '#options' => [
            'attributes' => [
              'class' => [
                'help-modal',
              ],
            ],
          ],
        ],
      ],
      'ga_check_results' => [
        '#type' => 'container',
        '#attributes' => [
          'class' => [
            'ga-check-results',
          ],
        ],
        'link' => [
          '#title' => t('How to check the results in Google Analytics'),
          '#type' => 'link',
          '#url' => Url::fromUri('base:/' . $this->abtestuiPath . '/help/check-ga-results.html'),
          '#options' => [
            'attributes' => [
              'class' => [
                'help-modal',
              ],
            ],
          ],
        ],
      ],
      'ga_account' => [
        '#markup' => ('<span class="ga-info"><span class="label">'
          . t('Google Analytics ID:')
          . ' </span><span class="ga-value">' . $ga_account_id . '</span></span>'),
      ],
    ];

    $output['table'] = [
      '#type' => 'table',
      '#header' => NULL,
      '#rows' => $rows,
      '#attributes' => [
        'class' => [
          'ab-test-table',
        ],
      ],
    ];

    $output['add_bottom'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => [
          'clearfix',
          'action-bar',
          'last',
        ],
      ],
      'link' => [
        '#title' => t('+ Add new test'),
        '#type' => 'link',
        '#url' => Url::fromRoute('abtestui.test_add_form'),
        '#options' => [
          'attributes' => [
            'class' => [
              'add-test-link',
              'last',
            ],
          ],
        ],
      ],
    ];

    // Add library.
    $output['#attached']['library'][] = 'abtestui/admin_design';

    return $output;
  }

}
